@@ -47,16 +47,29 @@
 #include "gsd-media-keys-manager-glue.h"
 
 #include "eggaccelerators.h"
-#include "actions/acme.h"
-#include "actions/acme-volume.h"
+#include "acme.h"
 #include "gsd-media-keys-window.h"
 
+#ifdef HAVE_PULSE
+#include <canberra-gtk.h>
+#include "gvc-mixer-control.h"
+#endif /* HAVE_PULSE */
+
 #define GSD_DBUS_PATH "/org/gnome/SettingsDaemon"
 #define GSD_DBUS_NAME "org.gnome.SettingsDaemon"
 #define GSD_MEDIA_KEYS_DBUS_PATH GSD_DBUS_PATH "/MediaKeys"
 #define GSD_MEDIA_KEYS_DBUS_NAME GSD_DBUS_NAME ".MediaKeys"
 
+#define TOUCHPAD_ENABLED_KEY "/desktop/gnome/peripherals/touchpad/touchpad_enabled"
+
 #define VOLUME_STEP 6           /* percents for one volume button press */
+#define MAX_VOLUME 65536.0
+
+#if defined(__OpenBSD__)
+# define EJECT_COMMAND "eject -t /dev/rcd0c"
+#else
+# define EJECT_COMMAND "eject -T"
+#endif
 
 #define GSD_MEDIA_KEYS_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GSD_TYPE_MEDIA_KEYS_MANAGER, GsdMediaKeysManagerPrivate))
 
@@ -67,7 +80,11 @@ typedef struct {
 
 struct GsdMediaKeysManagerPrivate
 {
-        AcmeVolume      *volume;
+#ifdef HAVE_PULSE
+        /* Volume bits */
+        GvcMixerControl *volume;
+        GvcMixerStream  *stream;
+#endif /* HAVE_PULSE */
         GtkWidget       *dialog;
         GConfClient     *conf_client;
 
@@ -246,7 +263,7 @@ static void
 dialog_init (GsdMediaKeysManager *manager)
 {
         if (manager->priv->dialog != NULL
-            && !gsd_media_keys_window_is_valid (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog))) {
+            && !gsd_osd_window_is_valid (GSD_OSD_WINDOW (manager->priv->dialog))) {
                 gtk_widget_destroy (manager->priv->dialog);
                 manager->priv->dialog = NULL;
         }
@@ -276,9 +293,12 @@ update_kbd_cb (GConfClient         *client,
                GsdMediaKeysManager *manager)
 {
         int      i;
+        gboolean need_flush = TRUE;
 
         g_return_if_fail (entry->key != NULL);
 
+        gdk_error_trap_push ();
+
         /* Find the key that was modified */
         for (i = 0; i < HANDLED_KEYS; i++) {
                 if (strcmp (entry->key, keys[i].gconf_key) == 0) {
@@ -286,7 +306,8 @@ update_kbd_cb (GConfClient         *client,
                         Key  *key;
 
                         if (keys[i].key != NULL) {
-                                grab_key (keys[i].key, FALSE, manager->priv->screens);
+                                need_flush = TRUE;
+                                grab_key_unsafe (keys[i].key, FALSE, manager->priv->screens);
                         }
 
                         g_free (keys[i].key);
@@ -301,14 +322,14 @@ update_kbd_cb (GConfClient         *client,
                         }
 
                         key = g_new0 (Key, 1);
-                        if (egg_accelerator_parse_virtual (tmp, &key->keysym, &key->keycode, &key->state) == FALSE
-                            || key->keycode == 0) {
+                        if (!egg_accelerator_parse_virtual (tmp, &key->keysym, &key->keycodes, &key->state)) {
                                 g_free (tmp);
                                 g_free (key);
                                 break;
                         }
 
-                        grab_key (key, TRUE, manager->priv->screens);
+                        need_flush = TRUE;
+                        grab_key_unsafe (key, TRUE, manager->priv->screens);
                         keys[i].key = key;
 
                         g_free (tmp);
@@ -316,15 +337,23 @@ update_kbd_cb (GConfClient         *client,
                         break;
                 }
         }
+
+        if (need_flush)
+                gdk_flush ();
+        if (gdk_error_trap_pop ())
+                g_warning ("Grab failed for some keys, another application may already have access the them.");
 }
 
 static void
 init_kbd (GsdMediaKeysManager *manager)
 {
         int i;
+        gboolean need_flush = FALSE;
 
         gnome_settings_profile_start (NULL);
 
+        gdk_error_trap_push ();
+
         for (i = 0; i < HANDLED_KEYS; i++) {
                 char *tmp;
                 Key  *key;
@@ -340,6 +369,7 @@ init_kbd (GsdMediaKeysManager *manager)
                 tmp = gconf_client_get_string (manager->priv->conf_client,
                                                keys[i].gconf_key,
                                                NULL);
+
                 if (!is_valid_shortcut (tmp)) {
                         g_debug ("Not a valid shortcut: '%s'", tmp);
                         g_free (tmp);
@@ -347,17 +377,8 @@ init_kbd (GsdMediaKeysManager *manager)
                 }
 
                 key = g_new0 (Key, 1);
-                if (!egg_accelerator_parse_virtual (tmp, &key->keysym, &key->keycode, &key->state)
-                    || key->keycode == 0) {
+                if (!egg_accelerator_parse_virtual (tmp, &key->keysym, &key->keycodes, &key->state)) {
                         g_debug ("Unable to parse: '%s'", tmp);
-
-                        g_free (tmp);
-                        g_free (key);
-                        continue;
-                }
-                /* avoid grabbing all the keyboard when KeyCode cannot be retrieved */
-                if (key->keycode == AnyKey)  {
-                        g_warning ("The shortcut key \"%s\" cannot be found on the current system, ignoring the binding", tmp);
                         g_free (tmp);
                         g_free (key);
                         continue;
@@ -367,9 +388,15 @@ init_kbd (GsdMediaKeysManager *manager)
 
                 keys[i].key = key;
 
-                grab_key (key, TRUE, manager->priv->screens);
+                need_flush = TRUE;
+                grab_key_unsafe (key, TRUE, manager->priv->screens);
         }
 
+        if (need_flush)
+                gdk_flush ();
+        if (gdk_error_trap_pop ())
+                g_warning ("Grab failed for some keys, another application may already have access the them.");
+
         gnome_settings_profile_end (NULL);
 }
 
@@ -556,7 +583,7 @@ do_www_action (GsdMediaKeysManager *manager,
 static void
 do_exit_action (GsdMediaKeysManager *manager)
 {
-        execute (manager, "gnome-session-save --kill", FALSE, FALSE);
+        execute (manager, "gnome-session-save --shutdown-dialog", FALSE, FALSE);
 }
 
 static void
@@ -565,8 +592,9 @@ do_eject_action (GsdMediaKeysManager *manager)
         char *command;
 
         dialog_init (manager);
-        gsd_media_keys_window_set_action (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
-                                          GSD_MEDIA_KEYS_WINDOW_ACTION_EJECT);
+        gsd_media_keys_window_set_action_custom (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
+                                                 "media-eject",
+                                                 FALSE);
         dialog_show (manager);
 
         command = gconf_client_get_string (manager->priv->conf_client,
@@ -575,81 +603,159 @@ do_eject_action (GsdMediaKeysManager *manager)
         if ((command != NULL) && (strcmp (command, "") != 0)) {
                 execute (manager, command, FALSE, FALSE);
         } else {
-                execute (manager, "eject -T", FALSE, FALSE);
+                execute (manager, EJECT_COMMAND, FALSE, FALSE);
         }
 
         g_free (command);
 }
 
+static void
+do_touchpad_action (GsdMediaKeysManager *manager)
+{
+        GConfClient *client = manager->priv->conf_client;
+        gboolean state = gconf_client_get_bool (client, TOUCHPAD_ENABLED_KEY, NULL);
+
+        dialog_init (manager);
+        gsd_media_keys_window_set_action_custom (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
+                                                 (!state) ? "touchpad-enabled" : "touchpad-disabled",
+                                                 FALSE);
+        dialog_show (manager);
+
+        gconf_client_set_bool (client, TOUCHPAD_ENABLED_KEY, !state, NULL);
+}
+
+#ifdef HAVE_PULSE
+static void
+update_dialog (GsdMediaKeysManager *manager,
+               guint vol,
+               gboolean muted)
+{
+        vol = (int) (100 * (double) vol / PA_VOLUME_NORM);
+        vol = CLAMP (vol, 0, 100);
+
+        dialog_init (manager);
+        gsd_media_keys_window_set_volume_muted (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
+                                                muted);
+        gsd_media_keys_window_set_volume_level (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog), vol);
+        gsd_media_keys_window_set_action (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
+                                          GSD_MEDIA_KEYS_WINDOW_ACTION_VOLUME);
+        dialog_show (manager);
+
+        ca_gtk_play_for_widget (manager->priv->dialog, 0,
+                                CA_PROP_EVENT_ID, "audio-volume-change",
+                                CA_PROP_EVENT_DESCRIPTION, "volume changed through key press",
+                                CA_PROP_APPLICATION_ID, "org.gnome.VolumeControl",
+                                NULL);
+}
+
 static void
 do_sound_action (GsdMediaKeysManager *manager,
                  int                  type)
 {
         gboolean muted;
-        int      vol;
-        int      vol_step;
-        GError  *error = NULL;
+        guint vol, norm_vol_step;
+        int vol_step;
 
-        if (manager->priv->volume == NULL) {
+        if (manager->priv->stream == NULL)
                 return;
-        }
 
         vol_step = gconf_client_get_int (manager->priv->conf_client,
                                          GCONF_MISC_DIR "/volume_step",
-                                         &error);
+                                         NULL);
 
-        if (error) {
+        if (vol_step <= 0 || vol_step > 100)
                 vol_step = VOLUME_STEP;
-                g_error_free (error);
-        }
 
-        if (vol_step > 0) {
-                int threshold = acme_volume_get_threshold (manager->priv->volume);
-                if (vol_step < threshold)
-                        vol_step = threshold;
-        }
+        norm_vol_step = PA_VOLUME_NORM * vol_step / 100;
 
         /* FIXME: this is racy */
-        vol = acme_volume_get_volume (manager->priv->volume);
-        muted = acme_volume_get_mute (manager->priv->volume);
+        vol = gvc_mixer_stream_get_volume (manager->priv->stream);
+        muted = gvc_mixer_stream_get_is_muted (manager->priv->stream);
 
         switch (type) {
         case MUTE_KEY:
-                acme_volume_mute_toggle (manager->priv->volume);
+                muted = !muted;
+                gvc_mixer_stream_change_is_muted (manager->priv->stream, muted);
                 break;
         case VOLUME_DOWN_KEY:
-                if (!muted && (vol <= vol_step)) {
-                        acme_volume_mute_toggle (manager->priv->volume);
+                if (!muted && (vol <= norm_vol_step)) {
+                        muted = !muted;
+                        vol = 0;
+                        gvc_mixer_stream_change_is_muted (manager->priv->stream, muted);
+                        if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE)
+                                gvc_mixer_stream_push_volume (manager->priv->stream);
+                } else if (!muted) {
+                        vol = vol - norm_vol_step;
+                        if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE)
+                                gvc_mixer_stream_push_volume (manager->priv->stream);
                 }
-                acme_volume_set_volume (manager->priv->volume, vol - vol_step);
                 break;
         case VOLUME_UP_KEY:
                 if (muted) {
+                        muted = !muted;
                         if (vol == 0) {
-                                acme_volume_set_volume (manager->priv->volume, vol + vol_step);
+                               vol = vol + norm_vol_step;
+                               gvc_mixer_stream_change_is_muted (manager->priv->stream, muted);
+                               if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE)
+                                        gvc_mixer_stream_push_volume (manager->priv->stream);
+                        } else {
+                                gvc_mixer_stream_change_is_muted (manager->priv->stream, muted);
                         }
-                        acme_volume_mute_toggle (manager->priv->volume);
                 } else {
-                        acme_volume_set_volume (manager->priv->volume, vol + vol_step);
+                        if (vol < MAX_VOLUME) {
+                                if (vol + norm_vol_step >= MAX_VOLUME) {
+                                        vol = MAX_VOLUME;
+                                } else {
+                                        vol = vol + norm_vol_step;
+                                }
+                                if (gvc_mixer_stream_set_volume (manager->priv->stream, vol) != FALSE)
+                                        gvc_mixer_stream_push_volume (manager->priv->stream);
+                        }
                 }
                 break;
         }
 
-        muted = acme_volume_get_mute (manager->priv->volume);
-        vol = acme_volume_get_volume (manager->priv->volume);
-
-        /* FIXME: AcmeVolume should probably emit signals
-           instead of doing it like this */
-        dialog_init (manager);
-        gsd_media_keys_window_set_volume_muted (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
-                                                muted);
-        gsd_media_keys_window_set_volume_level (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
-                                                vol);
-        gsd_media_keys_window_set_action (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
-                                          GSD_MEDIA_KEYS_WINDOW_ACTION_VOLUME);
-        dialog_show (manager);
+        update_dialog (manager, vol, muted);
 }
 
+static void
+update_default_sink (GsdMediaKeysManager *manager)
+{
+        GvcMixerStream *stream;
+
+        stream = gvc_mixer_control_get_default_sink (manager->priv->volume);
+        if (stream == manager->priv->stream)
+                return;
+
+        if (manager->priv->stream != NULL) {
+                g_object_unref (manager->priv->stream);
+                manager->priv->stream = NULL;
+        }
+
+        if (stream != NULL) {
+                manager->priv->stream = g_object_ref (stream);
+        } else {
+                g_warning ("Unable to get default sink");
+        }
+}
+
+static void
+on_control_ready (GvcMixerControl     *control,
+                  GsdMediaKeysManager *manager)
+{
+        update_default_sink (manager);
+}
+
+static void
+on_control_default_sink_changed (GvcMixerControl     *control,
+                                 guint                id,
+                                 GsdMediaKeysManager *manager)
+{
+        update_default_sink (manager);
+}
+
+#endif /* HAVE_PULSE */
+
 static gint
 find_by_application (gconstpointer a,
                      gconstpointer b)
@@ -661,10 +767,16 @@ static gint
 find_by_time (gconstpointer a,
               gconstpointer b)
 {
-        return ((MediaPlayer *)a)->time != 0 && ((MediaPlayer *)a)->time < ((MediaPlayer *)b)->time;
+        return ((MediaPlayer *)a)->time < ((MediaPlayer *)b)->time;
 }
 
-
+/*
+ * Register a new media player. Most applications will want to call
+ * this with time = GDK_CURRENT_TIME. This way, the last registered
+ * player will receive media events. In some cases, applications
+ * may want to register with a lower priority (usually 1), to grab
+ * events only nobody is interested.
+ */
 gboolean
 gsd_media_keys_manager_grab_media_player_keys (GsdMediaKeysManager *manager,
                                                const char          *application,
@@ -674,12 +786,19 @@ gsd_media_keys_manager_grab_media_player_keys (GsdMediaKeysManager *manager,
         GList       *iter;
         MediaPlayer *media_player;
 
+        if (time == GDK_CURRENT_TIME) {
+                GTimeVal tv;
+
+                g_get_current_time (&tv);
+                time = tv.tv_sec * 1000 + tv.tv_usec / 1000;
+        }
+
         iter = g_list_find_custom (manager->priv->media_players,
                                    application,
                                    find_by_application);
 
         if (iter != NULL) {
-                if (time == 0 || ((MediaPlayer *)iter->data)->time < time) {
+                if (((MediaPlayer *)iter->data)->time < time) {
                         g_free (((MediaPlayer *)iter->data)->application);
                         g_free (iter->data);
                         manager->priv->media_players = g_list_delete_link (manager->priv->media_players, iter);
@@ -688,6 +807,7 @@ gsd_media_keys_manager_grab_media_player_keys (GsdMediaKeysManager *manager,
                 }
         }
 
+        g_debug ("Registering %s at %u", application, time);
         media_player = g_new0 (MediaPlayer, 1);
         media_player->application = g_strdup (application);
         media_player->time = time;
@@ -711,6 +831,7 @@ gsd_media_keys_manager_release_media_player_keys (GsdMediaKeysManager *manager,
                                    find_by_application);
 
         if (iter != NULL) {
+                g_debug ("Deregistering %s", application);
                 g_free (((MediaPlayer *)iter->data)->application);
                 g_free (iter->data);
                 manager->priv->media_players = g_list_delete_link (manager->priv->media_players, iter);
@@ -752,10 +873,15 @@ do_action (GsdMediaKeysManager *manager,
         char *path;
 
         switch (type) {
+        case TOUCHPAD_KEY:
+                do_touchpad_action (manager);
+                break;
         case MUTE_KEY:
         case VOLUME_DOWN_KEY:
         case VOLUME_UP_KEY:
+#ifdef HAVE_PULSE
                 do_sound_action (manager, type);
+#endif /* HAVE_PULSE */
                 break;
         case POWER_KEY:
                 do_exit_action (manager);
@@ -784,9 +910,6 @@ do_action (GsdMediaKeysManager *manager,
         case EMAIL_KEY:
                 do_mail_action (manager);
                 break;
-        case SLEEP_KEY:
-                do_sleep_action ("apm", "xset dpms force off");
-                break;
         case SCREENSAVER_KEY:
                 if ((cmd = g_find_program_in_path ("gnome-screensaver-command"))) {
                         execute (manager, "gnome-screensaver-command --lock", FALSE, FALSE);
@@ -894,9 +1017,8 @@ acme_filter_events (GdkXEvent           *xevent,
         return GDK_FILTER_CONTINUE;
 }
 
-gboolean
-gsd_media_keys_manager_start (GsdMediaKeysManager *manager,
-                              GError             **error)
+static gboolean
+start_media_keys_idle_cb (GsdMediaKeysManager *manager)
 {
         GSList *l;
 
@@ -912,11 +1034,6 @@ gsd_media_keys_manager_start (GsdMediaKeysManager *manager,
         init_screens (manager);
         init_kbd (manager);
 
-        /* initialise Volume handler */
-        gnome_settings_profile_start ("acme_volume_new");
-        manager->priv->volume = acme_volume_new ();
-        gnome_settings_profile_end ("acme_volume_new");
-
         /* Start filtering the events */
         for (l = manager->priv->screens; l != NULL; l = l->next) {
                 gnome_settings_profile_start ("gdk_window_add_filter");
@@ -932,6 +1049,43 @@ gsd_media_keys_manager_start (GsdMediaKeysManager *manager,
 
         gnome_settings_profile_end (NULL);
 
+        return FALSE;
+}
+
+gboolean
+gsd_media_keys_manager_start (GsdMediaKeysManager *manager,
+                              GError             **error)
+{
+        gnome_settings_profile_start (NULL);
+
+#ifdef HAVE_PULSE
+        /* initialise Volume handler
+         *
+         * We do this one here to force checking gstreamer cache, etc.
+         * The rest (grabbing and setting the keys) can happen in an
+         * idle.
+         */
+        gnome_settings_profile_start ("gvc_mixer_control_new");
+
+        manager->priv->volume = gvc_mixer_control_new ("GNOME Volume Control Media Keys");
+
+        g_signal_connect (manager->priv->volume,
+                          "ready",
+                          G_CALLBACK (on_control_ready),
+                          manager);
+        g_signal_connect (manager->priv->volume,
+                          "default-sink-changed",
+                          G_CALLBACK (on_control_default_sink_changed),
+                          manager);
+
+        gvc_mixer_control_open (manager->priv->volume);
+
+        gnome_settings_profile_end ("gvc_mixer_control_new");
+#endif /* HAVE_PULSE */
+        g_idle_add ((GSourceFunc) start_media_keys_idle_cb, manager);
+
+        gnome_settings_profile_end (NULL);
+
         return TRUE;
 }
 
@@ -942,6 +1096,7 @@ gsd_media_keys_manager_stop (GsdMediaKeysManager *manager)
         GSList *ls;
         GList *l;
         int i;
+        gboolean need_flush;
 
         g_debug ("Stopping media_keys manager");
 
@@ -951,9 +1106,6 @@ gsd_media_keys_manager_stop (GsdMediaKeysManager *manager)
                                           manager);
         }
 
-        g_slist_free (priv->screens);
-        priv->screens = NULL;
-
         if (priv->conf_client) {
                 gconf_client_remove_dir (priv->conf_client,
                                          GCONF_BINDING_DIR,
@@ -975,15 +1127,38 @@ gsd_media_keys_manager_stop (GsdMediaKeysManager *manager)
                 priv->connection = NULL;
         }
 
+        need_flush = FALSE;
+        gdk_error_trap_push ();
+
         for (i = 0; i < HANDLED_KEYS; ++i) {
-                g_free (keys[i].key);
-                keys[i].key = NULL;
+                if (keys[i].key) {
+                        need_flush = TRUE;
+                        grab_key_unsafe (keys[i].key, FALSE, priv->screens);
+
+                        g_free (keys[i].key->keycodes);
+                        g_free (keys[i].key);
+                        keys[i].key = NULL;
+                }
+        }
+
+        if (need_flush)
+                gdk_flush ();
+        gdk_error_trap_pop ();
+
+        g_slist_free (priv->screens);
+        priv->screens = NULL;
+
+#ifdef HAVE_PULSE
+        if (priv->stream) {
+                g_object_unref (priv->stream);
+                priv->stream = NULL;
         }
 
         if (priv->volume) {
                 g_object_unref (priv->volume);
                 priv->volume = NULL;
         }
+#endif /* HAVE_PULSE */
 
         if (priv->dialog != NULL) {
                 gtk_widget_destroy (priv->dialog);
