@@ -45,10 +45,6 @@
 #include <libgnomeui/gnome-rr.h>
 #include <libgnomeui/gnome-rr-labeler.h>
 
-#ifdef HAVE_X11_EXTENSIONS_XRANDR_H
-#include <X11/extensions/Xrandr.h>
-#endif
-
 #ifdef HAVE_LIBNOTIFY
 #include <libnotify/notify.h>
 #endif
@@ -166,7 +162,9 @@ show_timestamps_dialog (GsdXrandrManager *manager, const char *msg)
 #endif
 }
 
-/* Optionally filters out GNOME_RR_ERROR_NO_MATCHING_CONFIG from
+/* This function centralizes the use of gnome_rr_config_apply_from_filename_with_time().
+ *
+ * Optionally filters out GNOME_RR_ERROR_NO_MATCHING_CONFIG from
  * gnome_rr_config_apply_from_filename_with_time(), since that is not usually an error.
  */
 static gboolean
@@ -206,6 +204,28 @@ fail:
         return FALSE;
 }
 
+/* This function centralizes the use of gnome_rr_config_apply_with_time().
+ *
+ * Applies a configuration and displays an error message if an error happens.
+ * We just return whether setting the configuration succeeded.
+ */
+static gboolean
+apply_configuration_and_display_error (GsdXrandrManager *manager, GnomeRRConfig *config, guint32 timestamp)
+{
+        GsdXrandrManagerPrivate *priv = manager->priv;
+        GError *error;
+        gboolean success;
+
+        error = NULL;
+        success = gnome_rr_config_apply_with_time (config, priv->rw_screen, timestamp, &error);
+        if (!success) {
+                error_message (manager, _("Could not switch the monitor configuration"), error, NULL);
+                g_error_free (error);
+        }
+
+        return success;
+}
+
 static void
 restore_backup_configuration_without_messages (const char *backup_filename, const char *intended_filename)
 {
@@ -468,7 +488,8 @@ is_laptop (GnomeOutputInfo *output)
         if (output->connected && output_name &&
             (strstr (output_name, "lvds")	||
              strstr (output_name, "LVDS")	||
-             strstr (output_name, "Lvds")))
+             strstr (output_name, "Lvds")       ||
+             strstr (output_name, "LCD")))
         {
                 return TRUE;
         }
@@ -673,12 +694,10 @@ make_laptop_setup (GnomeRRScreen *screen)
                 GnomeOutputInfo *info = result->outputs[i];
 
                 if (is_laptop (info)) {
-                        if (!info->on) {
-                                if (!turn_on (screen, info, 0, 0)) {
-                                        gnome_rr_config_free (result);
-                                        result = NULL;
-                                        break;
-                                }
+                        if (!turn_on (screen, info, 0, 0)) {
+                                gnome_rr_config_free (result);
+                                result = NULL;
+                                break;
                         }
                 }
                 else {
@@ -698,14 +717,8 @@ make_laptop_setup (GnomeRRScreen *screen)
 static int
 turn_on_and_get_rightmost_offset (GnomeRRScreen *screen, GnomeOutputInfo *info, int x)
 {
-        if (info->on) {
-                info->x = x;
-                info->y = 0;
+        if (turn_on (screen, info, x, 0))
                 x += info->width;
-        } else {
-                if (turn_on (screen, info, x, 0))
-                        x += info->width;
-        }
 
         return x;
 }
@@ -757,9 +770,8 @@ make_other_setup (GnomeRRScreen *screen)
                         info->on = FALSE;
                 }
                 else {
-                        if (info->connected && !info->on) {
+                        if (info->connected)
                                 turn_on (screen, info, 0, 0);
-                        }
                }
         }
 
@@ -995,11 +1007,7 @@ handle_fn_f7 (GsdXrandrManager *mgr, guint32 timestamp)
 
                 g_debug ("applying");
 
-                error = NULL;
-                if (!gnome_rr_config_apply_with_time (priv->fn_f7_configs[mgr->priv->current_fn_f7_config], screen, timestamp, &error)) {
-                        error_message (mgr, _("Could not switch the monitor configuration"), error, NULL);
-                        g_error_free (error);
-                }
+                apply_configuration_and_display_error (mgr, priv->fn_f7_configs[mgr->priv->current_fn_f7_config], timestamp);
         }
         else {
                 g_debug ("no configurations generated");
@@ -1077,7 +1085,6 @@ handle_rotate_windows (GsdXrandrManager *mgr, guint32 timestamp)
         GsdXrandrManagerPrivate *priv = mgr->priv;
         GnomeRRScreen *screen = priv->rw_screen;
         GnomeRRConfig *current;
-        GError *error;
         GnomeOutputInfo *rotatable_output_info;
         int num_allowed_rotations;
         GnomeRRRotation allowed_rotations;
@@ -1109,11 +1116,7 @@ handle_rotate_windows (GsdXrandrManager *mgr, guint32 timestamp)
 
         rotatable_output_info->rotation = next_rotation;
 
-        error = NULL;
-        if (!gnome_rr_config_apply_with_time (current, screen, timestamp, &error)) {
-                error_message (mgr, _("Could not switch the monitor configuration"), error, NULL);
-                g_error_free (error);
-        }
+        apply_configuration_and_display_error (mgr, current, timestamp);
 
 out:
         gnome_rr_config_free (current);
@@ -1275,13 +1278,8 @@ auto_configure_outputs (GsdXrandrManager *manager, guint32 timestamp)
 
         /* Apply the configuration! */
 
-        if (applicable) {
-                error = NULL;
-                if (!gnome_rr_config_apply_with_time (config, priv->rw_screen, timestamp, &error)) {
-                        error_message (manager, _("Could not switch the monitor configuration"), error, NULL);
-                        g_error_free (error);
-                }
-        }
+        if (applicable)
+                apply_configuration_and_display_error (manager, config, timestamp);
 
         g_list_free (just_turned_on);
         gnome_rr_config_free (config);
